//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////

// __AVR_ATmega328P__ , __AVR_ATmega168__
#define __AVR_ATmega328P__ 

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h> 
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>
#include <avr/wdt.h>

// include system util and driver headers
#include "CHAM_AVR_SYSTEM_V010.h"
#include "CHAM_AVR_UART_DRV_V010.h"
#include "CHAM_AVR_TWI_SPI_DRV_V010.h"
#include "CHAM_AVR_MOUSE_DRV_V010.h"
#include "CHAM_AVR_NTSC_DRV_V010.h"
#include "CHAM_AVR_VGA_DRV_V010.h"

// DEFINES AND MACROS /////////////////////////////////////////////////////////////////////////////////////////////////

// define CPU frequency in Mhz here if not defined in Makefile or other includes
#ifndef F_CPU
#define F_CPU 16000000UL // 28636360UL, 14318180UL, 21477270 UL
#endif

// set the define for the proper FLASH installed on Chameleon
#define AT26F004         4 // numeric codes for flash version
#define AT26DF081A       8

// set type of FLASH ROM currently installed in Chameleon (8MBit) for conditional compilation of code
#define FLASH_ROM        AT26DF081A // AT26F004, AT26DF081A

// Atmel 2F004/8xxx series basic commands
#define WRITE_ENABLE     0x06
#define WRITE_DISABLE    0x04
#define READ_STATUS      0x05
#define WRITE_STATUS     0x01
#define READ_DATA        0x03

#define SEQ_PROGRAM      0xAF  
#define BYTE_PROGRAM     0x02 

#define BLOCK_ERASE4     0x20
#define BLOCK_ERASE32    0x52
#define BLOCK_ERASE64    0xD8

#define SECTOR_ERASE     0x20    

#define CHIP_ERASE       0x60 
#define JEDEC_ID         0x9F
#define UNPROTECT_SECTOR 0x39 

// custom commands, not part of chip spec, just used here, start at 0xF0
#define UNPROTECT_CHIP   0xF0

// sound defines
#define SND_PORT      PORTD
#define SND_DDR       DDRD
#define SND_PIN       PD7

// servo defines
#define SERVO_PORT    PORTB
#define SERVO_DDR     DDRB
#define SERVO_PIN     PB3

// debugging constants for function terminal debug outputs
#define DEBUG_OFF     2    // no output at all
#define DEBUG_VERBOSE 1    // verbose mode for functions debug output
#define DEBUG_BRIEF   0    // brief mode for functions debug output

// ASCII codes for ease of parser development
#define ASCII_A       65
#define ASCII_0       48
#define ASCII_9       57
#define ASCII_BS      127  // backspace   
#define ASCII_LF      0x0A // line feed 
#define ASCII_CR      0x0D // carriage return
#define ASCII_ESC     0x1B // escape
#define ASCII_HEX     0x24 // $ for hex
#define ASCII_BIN     0x25 // % for binary
#define ASCII_LB      0x5B // [ 
#define ASCII_SEMI    0x3A // ; 
#define ASCII_EQUALS  0x3D // = 
#define ASCII_PERIOD  0x2E // .
#define ASCII_COMMA   0x2C // ,
#define ASCII_SHARP   0x23 // #
#define ASCII_NULL    0x00 // null character
#define ASCII_SPACE   0x20 // space

// conditional compilation controls, since AVR's FLASH is limited, set these to 0 to free up room

#define COMPILE_PORTIO_SUPPORT        0
#define COMPILE_HELP_SUPPORT          1
#define COMPILE_FLASH_SUPPORT         1 
#define COMPILE_RESET_SUPPORT         1
#define COMPILE_VGA_SUPPORT           0
#define COMPILE_NTSC_SUPPORT          0
#define COMPILE_PROP_REG_SUPPORT      0 
#define COMPILE_PROP_PORT_SUPPORT     0
#define COMPILE_GPU_GFX_SUPPORT       0
#define COMPILE_KEYBOARD_SUPPORT      0
#define COMPILE_MOUSE_SUPPORT         1
#define COMPILE_SOUND_SUPPORT         0
#define COMPILE_EEPROM_SUPPORT        0

// TYPES //////////////////////////////////////////////////////////////////////////////////////////////////////////////

// GLOBALS/////////////////////////////////////////////////////////////////////////////////////////////////////////////

// debuging values
int debug_wait_res_num_iter;
int debug_level = DEBUG_VERBOSE;

// flash string for testing
char *flash_string = "Atari 800 Rules Forever!!!";

// conditional compilation for FLASH memory variants
#if (FLASH_ROM == AT26F004)

// we could use PGM memory for these and or do it algorithmically, but this is cleaner for now
// 11 sectors [64K, 64K, 64K, 64K, 64K, 64K, 64K, 32K, 8K, 8k, 16K ]
//            |<------- (7) 64K sectors ------->|  1      2     1 

    #define NUM_FLASH_ROM_SECTORS 11

    long sector_addresses[] = { 0x00000,   // 64K sector 0
                                0x10000,   // 64K sector 1
                                0x20000,   // 64K sector 2
                                0x30000,   // 64K sector 3
                                0x40000,   // 64K sector 4
                                0x50000,   // 64K sector 5
                                0x60000,   // 64K sector 6
                                0x70000,   // 32K sector 7
                                0x78000,   // 8K sector 8  
                                0x7A000,   // 8K sector 9
                                0x7D000 }; // 8K sector 10
#endif


#if (FLASH_ROM == AT26DF081A)

// we could use PGM memory for these and or do it algorithmically, but this is cleaner for now
// 19 sectors [64K, 64K, 64K, ..., 64K, 64K, 64K, 64K, 16K, 8K, 8k, 32K ]
//            |<----------- (15) 64K sectors -------->| 1     2     1 

    #define NUM_FLASH_ROM_SECTORS 19

    long sector_addresses[] = { 0x00000,   // 64K sector 0 
                                0x10000,   // 64K sector 1
                                0x20000,   // 64K sector 2 
                                0x30000,   // 64K sector 3
                                0x40000,   // 64K sector 4
                                0x50000,   // 64K sector 5 
                                0x60000,   // 64K sector 6
                                0x70000,   // 64K sector 7 
                                0x80000,   // 64K sector 8
                                0x90000,   // 64K sector 9
                                0xA0000,   // 64K sector 10
                                0xB0000,   // 64K sector 11
                                0xC0000,   // 64K sector 12
                                0xD0000,   // 64K sector 13
                                0xE0000,   // 64K sector 14

                                0xF0000,   // 16K sector 15
                                0xF4000,   //  8K sector 16
                                0xF6000,   //  8K sector 17

                                0xF8000 }; // 32K sector 18
#endif



// EXTERNALS //////////////////////////////////////////////////////////////////////////////////////////////////////////

// PROTOTYPES /////////////////////////////////////////////////////////////////////////////////////////////////////////

// FUNCTIONS  /////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// UTILITY FUNCTIONS //////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INTERRUPT SERVICE ROUTINES /////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MAIN ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int main(void)
{

int x=16,y=10;                   // dummy vars for testing...

// variable below are all needed for the very crude command processor/shell, this is simply to show off the SPI commands
// for the default1/2 drivers and how you would send them WITHOUT the use of the API library.

// optimizer don't mess with the zohan!

volatile char sbuffer[80];       // general string buffer
volatile int  c;                 // input character
volatile long index;             // general looping var
volatile int  result;            // general result from function call
volatile char cmd_buff[64];      // command buffer, hold inputted command
volatile char token_buff[64];    // holds current token being processed
volatile int  num_tokens = 0;    // number of tokens found in input command string
volatile int  next_token = 0;    // next token being processed by algorithm
volatile char *tokens[8];        // storage for up to 8 tokens per command including command itself

volatile int  cmd_buff_index = -1;                // current token being processed
volatile long arg0, arg1, arg2, arg3, arg4, arg5; // generic args for command processor

volatile char *token_ptr = NULL;                  // alias pointer to token
volatile char *result_ptr;                        // alias pointer to result

volatile unsigned char spi_data = 0;              // spi data returned from sending a spi packet    

volatile unsigned char *eeprom_start_ptr, eeprom_value;    // eeprom access helper vars
volatile int eeprom_count;                                 // counting/indexing var for eeprom code

volatile long flash_start_addr, flash_value, flash_count;  // helper vars for flash code

volatile gid_event event;                                  // generic "event" for mouse, etc.

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INITIALIZATION   ///////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// initialize UART
UART_Init( UART_DEFAULT_BAUDRATE );
_delay_ms(100);

// enable interrupts
sei();

// clear terminal
UART_Transmit( 27 );
UART_Transmit( '[' );
UART_Transmit( '2' );
UART_Transmit( 'J' );

// home cursor
UART_Transmit( 27 );
UART_Transmit( '[' );
UART_Transmit( 'H' );

UART_Print_String("\nChameleon AVR Operating System (c) 2009 Nurve Networks LLC\n");

// set SPI rate
SPI_Init( SPI_DEFAULT_RATE );

// TOY COMMAND LINE INTERPRETER
#if 1

UART_Print_String("\nREADY>");

while(1)
{
if ( (c = UART_getc()) != -1 )
    {
    // special characters
    switch (c)
    {
    case 0x0D: // carriage return
    case 0x0A: // line feed
        {
        cmd_buff[ ++cmd_buff_index ] = 0;

        // any command to process?
        if (cmd_buff_index > 0)
           {        
            // process command...
            // first string better be command

            // upcase the string
            strupr( cmd_buff );

            // copy cmd_buffer to token_buffer for processing
            memcpy(token_buff, cmd_buff, strlen( cmd_buff ) + 1);

            // initialize command processor
            num_tokens = 0;

            // initialize token processor
            token_ptr = strtok_r( token_buff, "=, ", &result_ptr);
           
            while( token_ptr ) 
                 {
                 // insert token into token array and inc number of tokens found
                 tokens[ num_tokens++ ] = token_ptr;   
                 token_ptr = strtok_r( NULL, "=, ", &result_ptr);

                 } // end while

            // at this point we should have some tokens (first is command usually)
            if (num_tokens > 0)
               {
               // which command is user asking for?

               if ( !strcmp( tokens[0], "CLS") )
                  {
                  // enter handler
                  Terminal_Clear_Screen();
                  } // end if CLS


// TEST FIXTURE TO TEST API

               else
               if ( !strcmp( tokens[0], "TEST") )
                  {
                  // enter handler
                  // arg1 = atoi2( tokens[ 2 ] );


#if 1


                UART_Print_String("\nTesting FLASH Memory...\n");

                // STEP 1: enable writing to chip  ////////////////////////////////////////////////////////
                // enable SPI interface
   
                // set CS to SPI select channel 3 (FLASH)
                SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                UART_Print_String("\nExec Write Enable...");
                spi_data = SPI_WriteRead( WRITE_ENABLE );

                // disable SPI interface
   
                // set CS to SPI select channel 0 (null)
                SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                _delay_ms( 100 );

                // STEP 2: unprotect sector starting at address 0 //////////////////////////////////////////
                flash_start_addr = 0; // default to sector 0

                // enable SPI interface

                // set CS to SPI select channel 3 (FLASH)
                SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                UART_Print_String("\nExec Unprotect Sec Cmd...");
                spi_data = SPI_WriteRead( UNPROTECT_SECTOR );

                // write sector address, there are only a handful of sectors since each is 4-64K
                spi_data = SPI_WriteRead( flash_start_addr >> 16 );
                spi_data = SPI_WriteRead( flash_start_addr >> 8 );
                spi_data = SPI_WriteRead( flash_start_addr >> 0 );

                // disable SPI interface
                // set CS to SPI select channel 0 (null)
                SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                _delay_ms( 100 );

                // STEP 3: enable writing to chip again /////////////////////////////////////////////////////
                // enable SPI interface
   
                // set CS to SPI select channel 3 (FLASH)
                SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                UART_Print_String("\nExec Write Enable...");
                spi_data = SPI_WriteRead( WRITE_ENABLE );

                // disable SPI interface
   
                // set CS to SPI select channel 0 (null)
                SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                _delay_ms( 100 );

                // STEP 4: write test string starting at absolute address 0x00000 in flash //////////////////


                 // set CS to SPI select channel 3 (FLASH)
                 SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                 UART_Print_String("\nExec Seq Pgm...");

                 flash_start_addr = 0;
                 flash_count      = strlen( flash_string );
                 flash_value      = flash_string[ 0 ];


                 sprintf(sbuffer,"\n%ld %ld %s",flash_start_addr,
                                                flash_count,
                                                flash_string);


                 UART_Print_String( sbuffer );                            

                 // initiate seq program command
                 spi_data = SPI_WriteRead( SEQ_PROGRAM );

                 // write starting address byte by byte, MSB first...
                 spi_data = SPI_WriteRead( flash_start_addr >> 16 );
                 spi_data = SPI_WriteRead( flash_start_addr >> 8 );
                 spi_data = SPI_WriteRead( flash_start_addr >> 0 );

                 sprintf(sbuffer,"\nWrt string '%s' from [%ld..%ld]",flash_string, 
                                                                     flash_start_addr, 
                                                                     flash_start_addr + flash_count-1);
                                                             
                 UART_Print_String( sbuffer );                            

                 // write first byte
                 spi_data = SPI_WriteRead( flash_string[ 0 ] );

                 // de-assert set CS to SPI select channel 0 (null)
                 SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                 
                 // now write remaining data
                 for (index = 1; index < flash_count; index++)
                     {

                     // set CS to SPI select channel 3 (FLASH)
                     SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);
          
                     // re-issue seq program command, but NO address needed
                     spi_data = SPI_WriteRead( SEQ_PROGRAM );
                     // now data
                     spi_data = SPI_WriteRead( flash_string[ index ] );

                     // set CS to SPI select channel 0 (null)
                     SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                     // progress bar..    
                     UART_Transmit('.');

                     } // end for index

                // disable SPI interface
                
                // set CS to SPI select channel 0 (null)
                SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                // finally send write disable command

                // set CS to SPI select channel 3 (FLASH)
                SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);
                
                spi_data = SPI_WriteRead( WRITE_DISABLE );
                
                // set CS to SPI select channel 0 (null)
                SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

#endif

              

#if 1
                  UART_Print_String("\nPress <ENTER> to Continue PS/2 Mouse Test.");
                  while(UART_getc()==-1);

                  UART_Print_String("\nTesting Mouse & SPI Interace, press <ENTER> to Exit...");
                                 
                  while(UART_getc()==-1)
                       {
                       Mouse_Read(&event); // Keyboard_GotKey(); // Keyboard_State(0xC0); // Keyboard_Key();

                       event.x = event.x / 10;
                       event.y = event.y / 10;

                       if (event.buttons == 0X01)
                         {
                          NTSC_ClearScreen();
                          VGA_ClearScreen();
                          }
                  
                       if (event.x < -2)
                         event.x = -2;
                       else 
                       if (event.x > 2)
                           event.x = 2;

                       if (event.y < -2)
                         event.y = -2;
                       else 
                       if (event.y > 2)
                           event.y = 2;

                       x += event.x;
                       y -= event.y;

                       NTSC_SetXY( x, y );
                       NTSC_Term_Char('X');

                       VGA_SetXY( x, y );
                       VGA_Term_Char('X');

                       //sprintf(sbuffer,"\nx=%d, y=%d, z=%d, buttons=%X", x, y, event.z, event.buttons );
                       //UART_Print_String( sbuffer );                            

                       _delay_ms( 5 );
                       } // end while
#endif
                  } // end if



#if (COMPILE_EEPROM_SUPPORT == 1)
                /*
                Exercises the LOCAL AVR EEPROM, no SPI commication sent to Propeller. Just for testing purposes

                EEPROM .W [starting address 0..EEPROM size-1, count 0..2047, value 0..255] - writes a number of bytes at the starting address.
                EEPROM .R [starting address 0..EEPROM size-1, count 0..2047] - reads a number of bytes from the starting address.
                */
               else
               if ( !strcmp( tokens[0], "EEPROM") || !strcmp( tokens[0], "EE"))
                  {
                  // enter handler
                  
                  // test for write command
                  if ( !strcmp( tokens[1], ".W") || !strcmp( tokens[1], "WRITE")  )
                     {
                     UART_Print_String("\nExecuting EEPROM Write Byte(s) Command...");

                     eeprom_start_ptr = (unsigned char *)atoi2( tokens[ 2 ] );
                     eeprom_count     = atoi2( tokens[ 3 ] ) - atoi2( tokens[ 2 ] );
                     eeprom_value     = atoi2( tokens[ 4 ] );

                     for (index = 0; index <= eeprom_count; index++)
                         {
                         eeprom_write_byte( eeprom_start_ptr++, eeprom_value );
                         if ((index % 16) == 0 )
                            UART_Transmit('.');

                         } // end for index

                     } // end if write
                  else
                  if ( !strcmp( tokens[1], ".R") || !strcmp( tokens[1], "READ")  )
                     {
                     UART_Print_String("\nExecuting EEPROM Read Byte(s) Command...");

                     eeprom_start_ptr = (unsigned char *)atoi2( tokens[ 2 ] );
                     eeprom_count     = atoi2( tokens[ 3 ] ) - atoi2( tokens[ 2 ] );
                     
                     for (index = 0; index <= eeprom_count; index++)
                         {
                         eeprom_value = eeprom_read_byte( eeprom_start_ptr++ );

                         sprintf(sbuffer,"%d,",eeprom_value);
                         UART_Print_String( sbuffer );                            

                         } // end for index

                     } // end if read
                  else
                      { 
                      // problem...
                      UART_Print_String( "\nERROR - EEPROM command missing parameters / command not recognized.");

                      } // end else
                      
                  } // end if EEPROM
#endif

#if (COMPILE_SOUND_SUPPORT==1)
                /*
                Sound driver commands, only exposes a fraction of what the driver can do

                SOUND .PLAY [channel 0..3, frequency 0..2000, duration 0..7] - plays a sound on the selected channel with specified frequency and duration,
                                                                               a duration of 0 means play forever.

                SOUND  .STOP [channel 0..3] - stops the sent channel.                                                                                
                SOUND  .STOPALL             - stops all channels.
                */

               else
               if ( !strcmp( tokens[0], "SOUND") || !strcmp( tokens[0], "SND"))
                  {
                  // case .PLAY
                  if (!strcmp( tokens[1], ".PLAY" ) )
                      {

                      // SND [channel 0..3] [freq 0..2K] [ duration 0..7 ]
                      // enter handler
                      
                      // TODO - add some error handling to this!

                      // channel 0..3
                      arg1 = atoi2( tokens[ 2 ] );

                      // frequency tp play
                      arg2 = atoi2( tokens[ 3 ] );

                      // duration
                      arg3 = atoi2( tokens[ 4 ] );

                      // extract lower 2-bits of channel, shift to upper 2-bits of byte, then extract lower 3-bits of duration,
                      // shift behind channel, then finally OR in upper 3-bits of frequency
                      arg4 = (arg1 << 6) | (arg3 << 3) | (arg2 >> 8);

                      // mask off upper 6-bits of frequency and send only lower 8-bits 
                      arg5 = (arg2 & 0xFF);

                      // final data format
                      // arg4 [cc.dddfff ] | arg5 [ ffffffff ]
                      // channel 2-bit (0..3) | duration 3-bit 0..7 | frequency 11-bit (0 turn off channel)
                      SPI_Prop_Send_Cmd( SND_CMD_PLAYSOUNDFM, (int)arg5, (int)arg4);

                      } // end if .play  

                  // case .STOP
                  else
                  if ( !strcmp( tokens[1], ".STOP" ) )                                  
                     {

                     // [channel 0..3]
                     // enter handler
            
                     // channel 0..3
                     arg1 = ( atoi2( tokens[ 2 ] ) % 4);
            
                     // issue command to sound core
                     if (arg1 >= 0)
                        SPI_Prop_Send_Cmd( SND_CMD_STOPSOUND, (int)arg1, 0);

                     } // end if .STOP


                  // case .STOPALL
                  else
                  if ( !strcmp( tokens[1], ".STOPALL" ) )                                    
                    SPI_Prop_Send_Cmd( SND_CMD_STOPALLSOUNDS, 0, 0);


                  } // end if sound
#endif


#if (COMPILE_MOUSE_SUPPORT==1)
                /*
                Sends commands to the mouse driver
                Note - you must have the mouse plugged in and reset the chameleon to make a good connection, then you must "start" the mouse
                       since the default driver is the keyboard.

                MOUSE .ABS     - Displays the mouse in absolute coordinates, press key in console to stop.
                MOUSE .DELTA   - Displays the mouse in delta coordinates.
                MOUSE .START   - Unloads the keyboard driver and loads the mouse driver.
                MOUSE .STOP    - Unloads the mouse driver.
                MOUSE .PRESENT - Detects if the mouse is plugged in and returns the type of mouse.
                */

               else
               if ( !strcmp( tokens[0], "MOUSE") )
                  {
                   // read the mouse on the slave side and print to NTSC terminal until
                   // user hits a key on console

#if 1

                  if ( !strcmp( tokens[1], ".ABS"))
                     { 
                      while(UART_getc()==-1)
                           {
                           // request mouse read from slave, driver takes time to respond...

                           // read absolute x position first
                           SPI_Prop_Send_Cmd( MOUSE_CMD_ABS_X, 0x00, 0x00); 
                           _delay_us(SPI_PROP_DELAY_LONG_US);
    
                           // now read data from SPI buffer
                           arg1 = SPI_Prop_Send_Cmd( READ_CMD, 0x00, 0x00); 
                           _delay_us(SPI_PROP_DELAY_LONG_US);

                           // read absolute y position 
                           SPI_Prop_Send_Cmd( MOUSE_CMD_ABS_Y, 0x00, 0x00); 
                           _delay_us(SPI_PROP_DELAY_LONG_US);
    
                           // now read data from SPI buffer
                           arg2 = SPI_Prop_Send_Cmd( READ_CMD, 0x00, 0x00); 
                           _delay_us(SPI_PROP_DELAY_LONG_US);

                           // read absolute z position 
                           SPI_Prop_Send_Cmd( MOUSE_CMD_ABS_Z, 0x00, 0x00); 
                           _delay_us(SPI_PROP_DELAY_LONG_US);
    
                           // now read data from SPI buffer
                           arg3 = SPI_Prop_Send_Cmd( READ_CMD, 0x00, 0x00); 
                           _delay_us(SPI_PROP_DELAY_LONG_US);

                           // print results
                           
                           sprintf(sbuffer, "\nABS[X, Y, Z] = [%d, %d, %d]", (int)arg1, (int)arg2, (int)arg3);
                           UART_Print_String( sbuffer );

                           _delay_ms(25);
                        
                           } // end while

                  } // end if .ABS
                  else
#endif

                  
                  if ( !strcmp( tokens[1], ".DELTA"))
                     { 
                      while(UART_getc()==-1)
                           {
                           // request mouse read from slave, driver takes time to respond...

                           // read absolute x position first
                           SPI_Prop_Send_Cmd( MOUSE_CMD_DELTA_X, 0x00, 0x00); 
                           _delay_us(SPI_PROP_DELAY_LONG_US);
    
                           // now read data from SPI buffer
                           arg1 = SPI_Prop_Send_Cmd( READ_CMD, 0x00, 0x00) & 0xFF; 
                           _delay_us(SPI_PROP_DELAY_LONG_US);


                           // read absolute y position 
                           SPI_Prop_Send_Cmd( MOUSE_CMD_DELTA_Y, 0x00, 0x00); 
                           _delay_us(SPI_PROP_DELAY_LONG_US);
    
                           // now read data from SPI buffer
                           arg2 = SPI_Prop_Send_Cmd( READ_CMD, 0x00, 0x00) & 0xFF; 
                           _delay_us(SPI_PROP_DELAY_LONG_US);


                           // read absolute z position 
                           SPI_Prop_Send_Cmd( MOUSE_CMD_DELTA_Z, 0x00, 0x00); 
                           _delay_us(SPI_PROP_DELAY_LONG_US);
    
                           // now read data from SPI buffer
                           arg3 = SPI_Prop_Send_Cmd( READ_CMD, 0x00, 0x00)  & 0xFF; 
                           _delay_us(SPI_PROP_DELAY_LONG_US);


                           // print results
                           
                           sprintf(sbuffer, "\nDELTA[X, Y, Z] = [%d, %d, %d]", (signed char)arg1, (signed char)arg2, (signed char)arg3);
                           UART_Print_String( sbuffer );

                           _delay_ms(25);
                        
                           } // end while

                    } // end if .ABS

                else 
                if ( !strcmp( tokens[1], ".START")) 
                   {
                   SPI_Prop_Send_Cmd(  MOUSE_CMD_START, 0x00, 0x00);
                   } // end if .START
                else
                if ( !strcmp( tokens[1], ".STOP")) 
                   {
                   SPI_Prop_Send_Cmd(  MOUSE_CMD_STOP, 0x00, 0x00);
                   } // end if .STOP
                else
                if ( !strcmp( tokens[1], ".PRESENT")) 
                   {
                   SPI_Prop_Send_Cmd(  MOUSE_CMD_PRESENT, 0x00, 0x00);
                   
                   _delay_ms(1);

                   // now read result
                   arg1 = SPI_Prop_Send_Cmd( READ_CMD, 0x00, 0x00) & 0xFF;
                   
                   sprintf(sbuffer, "\nPresent = %d", (int)arg1);
                   UART_Print_String( sbuffer );

                   } // end if .PRESENT

                  } // end if read keyboard

#endif


#if (COMPILE_KEYBOARD_SUPPORT==1)
            /*
            Keyboard driver support, remember to plug a keyboard into the Chameleon AND the local keyboard on the PC that is connected to the console
            only controls when the program stops when you hit a key.

            KBD .READ     - Displays keys from the remote keyboard, press console key to stop.
            KBD .START    - Loads the keyboard driver, you only need to do this if you have unloaded or started the mouse driver.
            KBD .STOP     - Unloads the keyboard driver.
            KBD .PRESENT  - Tests if the keyboard is present.
            */


               else
               if ( !strcmp( tokens[0], "KBD") )
                  {
                   // read the keyboard on the slave side and print to NTSC terminal until
                   // user hits a key on console

                  if ( !strcmp( tokens[1], ".READ"))
                     { 

                     UART_Print_String( "\nKeyboard output:\n" );

                      while(UART_getc()==-1)
                           {
                           // request keyboard read from slave, driver takes time to respond...
                           SPI_Prop_Send_Cmd(  KEY_CMD_KEY, 0x00, 0x00); 
                           _delay_us(SPI_PROP_DELAY_LONG_US);

                           // keyboard should have enough time to be read now, request key by performing general read back command
                           // spi system will send back data in "result" buffer 

                           if ((arg2 = SPI_Prop_Send_Cmd(  READ_CMD, 0x00, 0x00)) !=0 )
                              {
                              // only need lower 8-bits of result, this is byte code
                              arg2 = arg2 & 0xFF;

                              sprintf(sbuffer, "\nkey = %c", (int)(arg2 - 32));
                              UART_Print_String( sbuffer );

                              } // end if

                           _delay_us(SPI_PROP_DELAY_SHORT_US);
                        
                           } // end while

                   } // end if .READ
                else 
                if ( !strcmp( tokens[1], ".START")) 
                   {
                   SPI_Prop_Send_Cmd(  KEY_CMD_START, 0x00, 0x00);
                   } // end if .START
                else
                if ( !strcmp( tokens[1], ".STOP")) 
                   {
                   SPI_Prop_Send_Cmd(  KEY_CMD_STOP, 0x00, 0x00);
                   } // end if .STOP
                else
                if ( !strcmp( tokens[1], ".PRESENT")) 
                   {
                   SPI_Prop_Send_Cmd(  KEY_CMD_PRESENT, 0x00, 0x00);
                   
                   _delay_ms(1);

                   // now read result
                   arg1 = SPI_Prop_Send_Cmd( READ_CMD, 0x00, 0x00);
                   
                   sprintf(sbuffer, "\nPresent = %d", (int)arg1);
                   UART_Print_String( sbuffer );

                   } // end if .PRESENT

                  } // end if read keyboard

#endif

#if (COMPILE_GPU_GFX_SUPPORT==1)

            /*
            Used to test GFX commands with "gfx" version 2 of Chameleon driver
            GFX GPU command

            GFX .W reg value
            GFX .R reg

            */

            else
            if ( !strcmp( tokens[0], "GFX" ) )
               {
              // case .W "write register"
              if ( !strcmp( tokens[1], ".W" ))                                    
                 {
                 // extract 8-bit register to write
                 arg1 = atoi2( tokens[2] );

                 // extract 8/16-bit value to write
                 arg2 = atoi2( tokens[3] );
              
                 // send write command: cmd, low, high
                 SPI_Prop_Send_Cmd( GPU_GFX_BASE_ID + arg1, (arg2 & 0xFF), (arg2 >> 8));
                 } // end if
                
              // case .R "read register"
              else
              if ( !strcmp( tokens[1], ".R" ))                                    
                 {
              
                 // extract 8-bit register to write
                 arg1 = atoi2( tokens[2] );
              
                 // send read command: cmd, 0,0
                 SPI_Prop_Send_Cmd( GPU_GFX_BASE_ID + arg1, 0x00, 0x00);
    
                 // let prop get data ready
                 _delay_ms(10);
                       
                 // now retrieve results, note, if we read the results back too soon, the driver might not have time to 
                 // set them, so keep that in mind
                 arg2 = SPI_Prop_Send_Cmd( READ_CMD, 0x00, 0x00);
             
                  
                 // print results
                 sprintf(sbuffer, "\nGFX GPU Register data = [%d]", (int)arg2);
                 UART_Print_String( sbuffer );
          

                 } // end if
                 
              } // end if GFX GPU command

#endif


#if (COMPILE_PROP_PORT_SUPPORT)

            /*
            propeller local 8-bit port I/O commands

            PORT .D [0..255] - sets the 8-bit I/O pin directions for the port 1=output, 0=input
            PORT .R          - reads the 8-bit port pins, outputs are don't cares, so you must ignore them and mask off
            PORT .W [0..255] - writes the 8-bit port pins, port pins set to input ignore data and won't get to pins

            */

            else
            if ( !strcmp( tokens[0], "PORT" ) )
               {
              // case .W "write port"
              if ( !strcmp( tokens[1], ".W" ))                                    
                 {
                 // extract 8-bit value to write
                 arg1 = atoi2( tokens[2] ) % 256;
              
                 // send write port command to prop
                 SPI_Prop_Send_Cmd( PORT_CMD_WRITE, arg1, 0x00);
                 } // end if
                
              // case .R "read port"
              else
              if ( !strcmp( tokens[1], ".R" ))                                    
                 {
                 // print the port out until user hits local terminal key
                 //while(UART_getc()==-1)
                   //   {
                      // send read port command to prop
                      SPI_Prop_Send_Cmd( PORT_CMD_READ, 0x00, 0x00);
    
                      _delay_us(SPI_PROP_DELAY_LONG_US);
                           
                      // now retrieve results, note, if we read the results back too soon, the driver might not have time to 
                      // set them, so keep that in mind
                      arg2 = SPI_Prop_Send_Cmd( READ_CMD, 0x00, 0x00);
                 
                      // only need lower 8-bits of result, this is 8-bit byte port only
                      // arg2 = arg2 & 0xFF;
                      
                      // print results
                      sprintf(sbuffer, "\nPropeller port data = [%d]", (int)arg2);
                      UART_Print_String( sbuffer );
                     // } // end while

                 } // end if
              // case .D "set direction of port bits"
              else
              if ( !strcmp( tokens[1], ".D" ))                                    
                 {
                 // extract 8-bit value to write as direction, 1=output, 0=input, 11110000 = out, out, out, out, in, in, in, in
                 arg1 = atoi2( tokens[2] ) % 256;
              
                 // send setdir port command to prop
                 SPI_Prop_Send_Cmd( PORT_CMD_SETDIR, arg1, 0x00);
                 } // end if
                 
              } // end if port command

#endif

#if (COMPILE_PROP_REGIO_SUPPORT==1)

            /*
            propeller COG 0 (SPI driver COG) register access commands - BEWARE!

            REG .W32 [0..15]        - performs a 32-bit write to the addressed register [0..F] from the output register buffer         
            REG .R32 [0..15]        - performs a 32-bit read from the addressed register [0..F] and stores in the input register buffer       
            REG .WB [0..3] [0..255] - write byte 0..3 of output register g_reg_out_buffer.byte[  0..3  ]         
            REG .RB [0..3]          - read byte 0..3 of inout register g_reg_in_buffer.byte[  0..3  ]         
            */

            else
            if ( !strcmp( tokens[0], "REG" ) )
               {
              // case .WCOG "write COG SFR register from 32-bit output buffer"
              if ( !strcmp( tokens[1], ".W32" ))                                    
                 {
                 // performs a 32-bit write to the addressed register [0..F] from the output register buffer         

                 // extract register 0..15 to write 32-bit output buffer to
                 arg1 = atoi2( tokens[2] ) % 16; // registers 0..15

                 // send write COG register to prop
                 SPI_Prop_Send_Cmd( REG_CMD_WRITE, arg1, 0x00);
                 } // end if
             else
              // case .RCOG "reads COG SFR register into 32-bit input buffer"
              if ( !strcmp( tokens[1], ".R32" ))                                    
                 {
                 // performs a 32-bit read from the addressed register [0..F] and stores in the input register buffer       

                 // extract register 0..15 to read into 32-bit input buffer
                 arg1 = atoi2( tokens[2] ) % 16; // registers 0..15

                 // send read COG register to prop
                 SPI_Prop_Send_Cmd( REG_CMD_READ, arg1, 0x00);
                 } // end if
              else
              // case .WB "write register byte buffer register 0..3"
              if ( !strcmp( tokens[1], ".WB" ))                                    
                 {
                 // write byte 0..3 of output register g_reg_out_buffer.byte[  0..3  ]         

                 // extract byte to write 0..3 to in 32-bit output buffer
                 arg1 = atoi2( tokens[2] ) % 4; // byte buffer register 0..3

                 // extract 8-bit data to write 
                 arg2 = atoi2( tokens[3] ) % 256; 

                 // send write byte register command to prop
                 SPI_Prop_Send_Cmd( REG_CMD_WRITE_BYTE, arg1, arg2);
                 } // end if
             else
              // case .RB "read register byte buffer register 0..3"
              if ( !strcmp( tokens[1], ".RB" ))                                    
                 {
                 // read byte 0..3 of input register g_reg_in_buffer.byte[  0..3 ]
                 // this data is then placed into spi buffer for transport back to client                 

                 // extract byte to read 0..3 from in 32-bit input buffer
                 arg1 = atoi2( tokens[2] ) % 4; // byte buffer register 0..3

                 // send read byte register command to prop
                 SPI_Prop_Send_Cmd( REG_CMD_READ_BYTE, arg1, 0x00);

                 // now that data has been read into SPI buffer, pull it out
                 Delay_Clocks(10000);

                 // now retrieve results, note, if we read the results back too soon, the driver might not have time to 
                 // set them, so keep that in mind
                 arg2 = SPI_Prop_Send_Cmd( READ_CMD, 0x00, 0x00);
             
                 // only need lower 8-bits of result
                 arg2 = arg2 & 0xFF;

                 // print results
                 sprintf(sbuffer, "\nByte[%d] = %d", (int)arg1, (int)arg2);
                 UART_Print_String( sbuffer );

                 } // end if

            } // end if REGister command

#endif


#if (COMPILE_NTSC_SUPPORT)

            /*
            NTSC Commands - These commands are processed and issued to NTSC terminal core.
           
            ntsc .pr [string]      - This command prints a single string token to the NTSC terminal. 
            ntsc .cls              - Clear and home the NTSC terminal.
            ntsc .col [num 0..7]   - Set color of NTSC terminal
            ntsc .nl               - Print a newline to NTSC terminal.
            ntsc .out [num 0..255] - Outputs a single ASCII character to NTSC terminal.
            ntsc .sx [num 0..39]   - Sets X cursor position of NTSC terminal.
            ntsc .sy [num 0..29]   - Sets Y cursor position of NTSC terminal.
            ntsc .gx               - Gets X cursor position of NTSC terminal.
            ntsc .gy               - Gets Y cursor position of NTSC terminal.


            .OUT - Output a character supports these sub-commands
            
                 $00 = clear screen
                 $01 = home
                 $08 = backspace
                 $09 = tab (8 spaces per)
                 $0A = set X position (X follows)
                 $0B = set Y position (Y follows)
                 $0C = set color (color follows)
                 $0D = carriage return 
           */

            else
            if ( !strcmp( tokens[0], "NTSC" ) )
               {
              // case .PR "print"
              if ( !strcmp( tokens[1], ".PR" ))                                    
                 {

                 SPI_Prop_Print_String( DEVICE_NTSC, tokens[ 2 ] );
                 SPI_Prop_Send_Cmd( GFX_CMD_NTSC_PRINTCHAR, 0x0D, 0x00);
                 } // end if
                    
              // case .CLS "clear screen", note we could implement this command directly thru the OUT character
              // command, but then we can only use the out() function within the graphics driver, breaking the "clearscreen"
              // out as a command, allows us to do some extra processing on both sides of the SPI driver, locally and in the 
              // Prop SPI driver interface, in essence we are overloading the command to add some functionality to it, or in 
              // this case just doing it for fun as an example. 
              else
              if ( !strcmp( tokens[1], ".CLS" ) )
                 {
                 // clear screen                                     
                 SPI_Prop_Send_Cmd( GFX_CMD_NTSC_CLS, 0x00, 0x00);  
                _delay_ms(100);

#if 0
                 // font test
                 for (arg1 = 0; arg1 < 144; arg1++)
                     {
                     
                     if ( (arg1 % 16) == 0)
                        {
                        SPI_Prop_Send_Cmd( GFX_CMD_NTSC_PRINTCHAR, 0x0D, 0x00);
                        _delay_ms(1);
                        SPI_Prop_Send_Cmd( GFX_CMD_NTSC_PRINTCHAR, 32, 0x00);
                        _delay_ms(1);
                        SPI_Prop_Send_Cmd( GFX_CMD_NTSC_PRINTCHAR, 32, 0x00);
                        _delay_ms(1);
                        
                        } // end if

                     SPI_Prop_Send_Cmd( GFX_CMD_NTSC_PRINTCHAR, arg1+32, 0x00);
                     _delay_ms(1);
                   
                     } // end for arg1
#endif 

                 } // end if

              //  case .OUT "output a single character"
              else
              if ( !strcmp( tokens[1], ".OUT" ) || !strcmp( tokens[1], ".O" ) )                                    
                 {
                 arg1 = atoi2( tokens[2] ) % 256;
                 SPI_Prop_Send_Cmd( GFX_CMD_NTSC_PRINTCHAR, arg1, 0x00);
                 } // end if

              //  case .NL "newline"
              else
              if ( !strcmp( tokens[1], ".NL" ) )                                    
                {
                SPI_Prop_Send_Cmd( GFX_CMD_NTSC_PRINTCHAR, ASCII_CR, 0x00);
                } // end if

              // case .COL "set color"
              else
              if ( !strcmp( tokens[1], ".COL" ) )
                 {
                 // extract color 0..7
                 arg1 = atoi2( tokens[ 2 ]) % 8;
                
                 // send driver set color command, [$0C, col 0..7]                           
                 SPI_Prop_Send_Cmd( GFX_CMD_NTSC_PRINTCHAR, 0x0C, 0x00);
                 _delay_us(SPI_PROP_DELAY_SHORT_US);
                 SPI_Prop_Send_Cmd( GFX_CMD_NTSC_PRINTCHAR, arg1, 0x00);
                 } // end if

              // case .SX "set cursor x position"
              else
              if ( !strcmp( tokens[1], ".SX") )
                {
                // extract x
                arg1 = atoi2( tokens[ 2 ]) % 40;

                // send driver set x command, [$0A, x 0..39]           
                SPI_Prop_Send_Cmd( GFX_CMD_NTSC_PRINTCHAR, 0x0A, 0x00);
                 _delay_us(SPI_PROP_DELAY_SHORT_US);
                SPI_Prop_Send_Cmd( GFX_CMD_NTSC_PRINTCHAR, arg1, 0x00);
                } // end if

              // case .SY "set cursor y position"
              else
              if ( !strcmp( tokens[1], ".SY" ) )
                 {
                 // extract y
                 arg1 = atoi2( tokens[ 2 ]) % 30;

                 // send driver set x command, [$0B, y 0..29]                           
               
                 SPI_Prop_Send_Cmd( GFX_CMD_NTSC_PRINTCHAR, 0x0B, 0x00);
                 _delay_us(SPI_PROP_DELAY_SHORT_US);
                 SPI_Prop_Send_Cmd( GFX_CMD_NTSC_PRINTCHAR, arg1, 0x00);
                 } // end if

              // case .GX "get cursor x position"
              else
              if ( !strcmp( tokens[1], ".GX") )
                {
                // send driver get x command
                SPI_Prop_Send_Cmd( GFX_CMD_NTSC_GETX, 0x00, 0x00);

                _delay_us(SPI_PROP_DELAY_LONG_US);

                // now read the data back, note 0x00, 0x00 are dummy data, mask lower 8-bits
                arg1 = SPI_Prop_Send_Cmd( READ_CMD, 0x00, 0x00) & 0xFF;

                sprintf(sbuffer, "\nX-Pos / Col = %d", (int)arg1);
                UART_Print_String( sbuffer );

                } // end if

              // case .GY "get cursor x position"
              else
              if ( !strcmp( tokens[1], ".GY") )
                {
                // send driver get y command
                SPI_Prop_Send_Cmd( GFX_CMD_NTSC_GETY, 0x00, 0x00);

                _delay_us(SPI_PROP_DELAY_LONG_US);

                // now read the data back, note 0x00, 0x00 are dummy data, mask lower 8-bits
                arg1 = SPI_Prop_Send_Cmd( READ_CMD, 0x00, 0x00) & 0xFF;

                sprintf(sbuffer, "\nY-Pos / Row = %d",(int)arg1);
                UART_Print_String( sbuffer );

                } // end if


               } // end ntsc case

#endif

#if (COMPILE_VGA_SUPPORT == 1)

            /*
            VGA Commands - These commands are processed and issued to VGA terminal core.
           
            vga .pr [string]      - This command prints a single string token to the VGA terminal. 
            vga .cls              - Clear and home the VGA terminal.
            vga .col [num 0..7]   - Set color of VGA terminal
            vga .nl               - Print a newline to VGA terminal.
            vga .out [num 0..255] - Outputs a single ASCII character to VGA terminal.
            vga .sx [num 0..39]   - Sets X cursor position of VGA terminal.
            vga .sy [num 0..29]   - Sets Y cursor position of VGA terminal.
            vga .gx               - Gets X cursor position of VGA terminal.
            vga .gy               - Gets Y cursor position of VGA terminal.


            .OUT - Output a character supports these sub-commands
            
                 $00 = clear screen
                 $01 = home
                 $08 = backspace
                 $09 = tab (8 spaces per)
                 $0A = set X position (X follows)
                 $0B = set Y position (Y follows)
                 $0C = set color (color follows)
                 $0D = carriage return 
            */


            else
            if ( !strcmp( tokens[0], "VGA" ) )
               {
              // case .PR "print"
              if ( !strcmp( tokens[1], ".PR" ) || !strcmp( tokens[1], ".P" ))                                    
                 {

                 SPI_Prop_Print_String( DEVICE_VGA, tokens[ 2 ] );
                 SPI_Prop_Send_Cmd( GFX_CMD_VGA_PRINTCHAR, 0x0D, 0x00);
                 } // end if
                    
              // case .CLS "clear screen", note we could implement this command directly thru the OUT character
              // command, but then we can only use the out() function within the graphics driver, breaking the "clearscreen"
              // out as a command, allows us to do some extra processing on both sides of the SPI driver, locally and in the 
              // Prop SPI driver interface, in essence we are overloading the command to add some functionality to it, or in 
              // this case just doing it for fun as an example. 
              else
              if ( !strcmp( tokens[1], ".CLS" ) )
                 {
                 // clear screen                                     
                 SPI_Prop_Send_Cmd( GFX_CMD_VGA_CLS, 0x00, 0x00);  
                 } // end if

              //  case .OUT "output a single character"
              else
              if ( !strcmp( tokens[1], ".OUT" ) || !strcmp( tokens[1], ".O" ) )                                    
                 {
                 arg1 = atoi2( tokens[2] ) % 256;
                 SPI_Prop_Send_Cmd( GFX_CMD_VGA_PRINTCHAR, arg1, 0x00);
                 } // end if

              //  case .NL "newline"
              else
              if ( !strcmp( tokens[1], ".NL" ) )                                    
                {
                SPI_Prop_Send_Cmd( GFX_CMD_VGA_PRINTCHAR, ASCII_CR, 0x00);
                } // end if

              // case .COL "set color"
              else
              if ( !strcmp( tokens[1], ".COL" ) )
                 {
                 // extract color 0..7
                 arg1 = atoi2( tokens[ 2 ]) % 8;
                
                 // send driver set color command, [$0C, col 0..7]                           
                 SPI_Prop_Send_Cmd( GFX_CMD_VGA_PRINTCHAR, 0x0C, 0x00);
                 _delay_us(SPI_PROP_DELAY_SHORT_US);
                 SPI_Prop_Send_Cmd( GFX_CMD_VGA_PRINTCHAR, arg1, 0x00);
                 } // end if

              // case .SX "set cursor x position"
              else
              if ( !strcmp( tokens[1], ".SX") )
                {
                // extract x
                arg1 = atoi2( tokens[ 2 ]) % 32;

                // send driver set x command, [$0A, x 0..31]           
                SPI_Prop_Send_Cmd( GFX_CMD_VGA_PRINTCHAR, 0x0A, 0x00);
                 _delay_us(SPI_PROP_DELAY_SHORT_US);
                SPI_Prop_Send_Cmd( GFX_CMD_VGA_PRINTCHAR, arg1, 0x00);
                } // end if

              // case .SY "set cursor y position"
              else
              if ( !strcmp( tokens[1], ".SY" ) )
                 {
                 // extract y
                 arg1 = atoi2( tokens[ 2 ]) % 15;

                 // send driver set x command, [$0B, y 0..14]                           
               
                 SPI_Prop_Send_Cmd( GFX_CMD_VGA_PRINTCHAR, 0x0B, 0x00);
                 _delay_us(SPI_PROP_DELAY_SHORT_US);
                 SPI_Prop_Send_Cmd( GFX_CMD_VGA_PRINTCHAR, arg1, 0x00);
                 } // end if

              // case .GX "get cursor x position"
              else
              if ( !strcmp( tokens[1], ".GX") )
                {
                // send driver get x command
                SPI_Prop_Send_Cmd( GFX_CMD_VGA_GETX, 0x00, 0x00);

                _delay_us(SPI_PROP_DELAY_LONG_US);

                // now read the data back, note 0x00, 0x00 are dummy data, mask lower 8-bits
                arg1 = SPI_Prop_Send_Cmd( READ_CMD, 0x00, 0x00) & 0xFF;

                sprintf(sbuffer, "\nX-Pos / Col = %d", (int)arg1);
                UART_Print_String( sbuffer );

                } // end if

              // case .GY "get cursor x position"
              else
              if ( !strcmp( tokens[1], ".GY") )
                {
                // send driver get y command
                SPI_Prop_Send_Cmd( GFX_CMD_VGA_GETY, 0x00, 0x00);

                _delay_us(SPI_PROP_DELAY_LONG_US);

                // now read the data back, note 0x00, 0x00 are dummy data, mask lower 8-bits
                arg1 = SPI_Prop_Send_Cmd( READ_CMD, 0x00, 0x00) & 0xFF;

                sprintf(sbuffer, "\nY-Pos / Row = %d", (int)arg1);
                UART_Print_String( sbuffer );

                } // end if



               } // end vga case

#endif

#if (COMPILE_RESET_SUPPORT)
            /*
            Allows resetting of the LOCAL AVR and the REMOTE Propeller

            RESET .AVR  - Resets the AVR.
            RESET .PROP - Resets the Propeller.
            RESET .ALL  - Resets them both.
            */
               else
               if ( !strcmp( tokens[0], "RESET") )
                  {
                  // allows the AVR, Prop or both to be reset
                 if (!strcmp( tokens[1], ".AVR") )  
                    {
                    // reset the AVR
                    AVR_Soft_Reset();

                    } // end if
                 else
                 if (!strcmp( tokens[1], ".PROP") )  
                    {
                    // reset the prop
                    SPI_Prop_Send_Cmd( SYS_RESET, 0x00, 0x00);
                    } // end if
                 else
                 if (!strcmp( tokens[1], ".ALL") )  
                    {
                    // reset the prop
                    SPI_Prop_Send_Cmd( SYS_RESET, 0x00, 0x00);

                    // reset the AVR
                    AVR_Soft_Reset();
                  
                    } // end if

                  } // end if RESET

#endif

#if (COMPILE_FLASH_SUPPORT)

               else
               if ( !strcmp( tokens[0], "FLASH") || !strcmp( tokens[0], "FL"))
                  {
                  // enter handler
                  arg0 = 0; // null command    
                  next_token = 0;
                                   
                  // determine which command is being requested, parse first parameter, search for "cmd=" or a number

                  if (!strcmp( tokens[1], ".CMD"))
                     {
                     // user is entering command directly with a number in hex or decimal format
                     arg0 = atoi2( tokens[2] );
                     next_token = 3;
                     
                     //sprintf(sbuffer, "parsed: CMD = %d", arg0);                    
                     //UART_Print_String( sbuffer );

                     } // end if  
                  else if (!strcmp( tokens[1], ".WE"))  {arg0 = WRITE_ENABLE;  next_token = 2; }
                  else if (!strcmp( tokens[1], ".WD")) {arg0 = WRITE_DISABLE; next_token = 2; }
                  else if (!strcmp( tokens[1], ".RS") || !strcmp( tokens[1], "STATUS"))   {arg0 = READ_STATUS;   next_token = 2; }    
                  //else if (!strcmp( tokens[1], "WRITE_STATUS"))  {arg0 = WRITE_STATUS;  next_token = 2; }   
                  else if (!strcmp( tokens[1], ".RD"))     {arg0 = READ_DATA;     next_token = 2; }   
                  else if (!strcmp( tokens[1], ".SP") )  {arg0 = SEQ_PROGRAM;  next_token = 2; }   
                  else if (!strcmp( tokens[1], ".BP") )  {arg0 = BYTE_PROGRAM;  next_token = 2; }   
                  else if (!strcmp( tokens[1], ".BE4"))   {arg0 = BLOCK_ERASE4;   next_token = 2; }   
                  //else if (!strcmp( tokens[1], "SECTOR_ERASE"))  {arg0 = SECTOR_ERASE;  next_token = 2; }       
                  else if (!strcmp( tokens[1], ".CE"))    {arg0 = CHIP_ERASE;    next_token = 2; }
                  else if (!strcmp( tokens[1], ".JI"))      {arg0 = JEDEC_ID;      next_token = 2; }
                  else if (!strcmp( tokens[1], ".US"))  {arg0 = UNPROTECT_SECTOR;  next_token = 2; }
                  else if (!strcmp( tokens[1], ".UC"))  {arg0 = UNPROTECT_CHIP;  next_token = 2; }
                  else
                     {
                     UART_Print_String( "\nFLASH parms wrong");
                     } // end else malformed command


                   //sprintf(sbuffer, "\nparsed: CMD = %d", arg0);                    
                   //UART_Print_String( sbuffer );

                   // what is the command?
                   switch( arg0 )
                   {

#if 1

                   case UNPROTECT_CHIP:
                        {

                        // unprotect the chip sector by sector, default on power up is protected, pain, but we have to do this...

                        UART_Print_String("\nExec Unprotect chip...");

                        // step 1: send WRITE_ENABLE command

                        // set CS to SPI select channel 3 (FLASH)
                        SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);
                        spi_data = SPI_WriteRead( WRITE_ENABLE );

                        // disable SPI interface
                        // set CS to SPI select channel 0 (null)
                        SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                        // step 2: unprotect all sectors from sector address list
                        for (index = 0; index < NUM_FLASH_ROM_SECTORS; index++)
                            {
                            // enable SPI interface
                            // set CS to SPI select channel 3 (FLASH)
                            SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                            spi_data = SPI_WriteRead( UNPROTECT_SECTOR );

                            // write sector address, there are only a handful of sectors since each is 4-64K
                            spi_data = SPI_WriteRead( sector_addresses[ index ] >> 16 );
                            spi_data = SPI_WriteRead( sector_addresses[ index ] >> 8 );
                            spi_data = SPI_WriteRead( sector_addresses[ index ] >> 0 );

                            // disable SPI interface
                            // set CS to SPI select channel 0 (null)
                            SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                            // progress bar..    
                            UART_Transmit('.');

                            _delay_s(1);

                            } // end for index

                        } break;

#endif

#if 1

                    case UNPROTECT_SECTOR:
                        {
                        // get address for sector to unprotect
                        if ( (num_tokens-1) >= next_token )
                          flash_start_addr = atoi2( tokens[ next_token ] );
                        else
                          flash_start_addr = 0; // default to sector 0

                        // enable SPI interface

                        // set CS to SPI select channel 3 (FLASH)
                        SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                        UART_Print_String("\nExec Unprotect Sec Cmd...");
                        spi_data = SPI_WriteRead( UNPROTECT_SECTOR );

                        // write sector address, there are only a handful of sectors since each is 4-64K
                        spi_data = SPI_WriteRead( flash_start_addr >> 16 );
                        spi_data = SPI_WriteRead( flash_start_addr >> 8 );
                        spi_data = SPI_WriteRead( flash_start_addr >> 0 );

                        // disable SPI interface
                        // set CS to SPI select channel 0 (null)
                        SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                        } break;


#endif

#if 1

                    case WRITE_ENABLE:
                        {
                        // enable SPI interface
           
                        // set CS to SPI select channel 3 (FLASH)
                        SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                        UART_Print_String("\nExec Write Enable...");
                        spi_data = SPI_WriteRead( WRITE_ENABLE );

                        // disable SPI interface
           
                        // set CS to SPI select channel 0 (null)
                        SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                        } break;
#endif

#if 1
                    case WRITE_DISABLE: 
                        {
                        // enable SPI interface
                        //RESET_BIT(SPI_PORT, SPI_CS);
                        // set CS to SPI select channel 3 (FLASH)
                        SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                        UART_Print_String("\nExec Write Disable...");
                        spi_data = SPI_WriteRead( WRITE_DISABLE );

                        // disable SPI interface
                        //SET_BIT(SPI_PORT, SPI_CS);

                        // set CS to SPI select channel 0 (null)
                        SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);
                        } break;

#endif

#if 1
                    case READ_STATUS:     
                        {
                        // check for iterations command, any tokens left?
                        if ( (num_tokens-1) >= next_token )
                          arg1 = atoi2( tokens[ next_token ] );
                        else
                          arg1 = 1;

                        // enable SPI interface
                        //RESET_BIT(SPI_PORT, SPI_CS);
                        // set CS to SPI select channel 3 (FLASH)
                        SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                        UART_Print_String("\nExec Read Stat Reg...");
                        spi_data = SPI_WriteRead( READ_STATUS );
                  
                        for (index = 0; index < arg1; index++)
                            {
                            spi_data = SPI_WriteRead( 0xFF );

                            sprintf(sbuffer, "\nStatus = %d", spi_data);                    
                            UART_Print_String( sbuffer );

                            } // end for

                        // disable SPI interface
                        //SET_BIT(SPI_PORT, SPI_CS);
                        // set CS to SPI select channel 0 (null)
                        SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                        } break;
#endif


#if 0
                    case WRITE_STATUS:   
                        {

                        // check for iterations command, any tokens left?
                        if ( (num_tokens-1) >= next_token )
                          arg1 = atoi2( tokens[ next_token ] );
                        else
                          arg1 = 0;

                        // enable SPI interface
                        //RESET_BIT(SPI_PORT, SPI_CS);
                        // set CS to SPI select channel 3 (FLASH)
                        SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                        UART_Print_String("\nExec Write Status Reg...");
                        spi_data = SPI_WriteRead( WRITE_STATUS );
                  
                        // disable SPI interface
                        //SET_BIT(SPI_PORT, SPI_CS);
                        // set CS to SPI select channel 0 (null)
                        SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);
                        } break;
#endif


#if 1
                    case JEDEC_ID:
                        {
                        //RESET_BIT(SPI_PORT, SPI_CS);
                        // set CS to SPI select channel 3 (FLASH)
                        SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                        UART_Print_String("\nExec JEDEC ID...");

                        spi_data = SPI_WriteRead( JEDEC_ID );
                    
                        // now read the id data back
                        
                        sprintf(sbuffer,"\nManf = %d,", SPI_WriteRead( 0x00 ) );
                        UART_Print_String( sbuffer );                            
                                                      
                        sprintf(sbuffer,"\nMem Type = %d,", SPI_WriteRead( 0x00 ) );
                        UART_Print_String( sbuffer );                            

                        sprintf(sbuffer,"\nCap = %d,", SPI_WriteRead( 0x00 ) );
                        UART_Print_String( sbuffer );                            

                        // disable SPI interface
                        //SET_BIT(SPI_PORT, SPI_CS);
                        // set CS to SPI select channel 0 (null)
                        SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                        } break;
#endif

#if 1

                    case READ_DATA:       
                        {
                         //RESET_BIT(SPI_PORT, SPI_CS);

                         // set CS to SPI select channel 3 (FLASH)
                         SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                         UART_Print_String("\nExec Read Byte(s)...");

                         flash_start_addr = atoi2( tokens[ 2 ] );
                         flash_count      = atoi2( tokens[ 3 ] ); 
                     
                         // initiate read command
                         spi_data = SPI_WriteRead( READ_DATA );

                         // write starting address byte by byte, MSB first...
                         spi_data = SPI_WriteRead( flash_start_addr >> 16 );
                         spi_data = SPI_WriteRead( flash_start_addr >> 8 );
                         spi_data = SPI_WriteRead( flash_start_addr >> 0 );

                         // now read data 
                         for (index = 0; index < flash_count; index++)
                             {
                             flash_value = SPI_WriteRead( 0x00 );

                             sprintf(sbuffer,"%c,",(char)flash_value);
                             UART_Print_String( sbuffer );                            

                             } // end for index

                        // disable SPI interface
                        //SET_BIT(SPI_PORT, SPI_CS);
                        // set CS to SPI select channel 0 (null)
                        SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);
                        
                        } break;
#endif

#if 1
                    case SEQ_PROGRAM: 
                        {

                        // set CS to SPI select channel 3 (FLASH)
                        SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                         UART_Print_String("\nExec Seq Pgm...");

                         flash_start_addr = atoi2( tokens[ 2 ] );
                         flash_count      = atoi2( tokens[ 3 ] ); 
                         flash_value      = atoi2( tokens[ 4 ] ); 


                         sprintf(sbuffer,"\n%ld %ld %ld",flash_start_addr,
                                                       flash_count,
                                                       flash_value);


                         UART_Print_String( sbuffer );                            

                         // initiate seq program command
                         spi_data = SPI_WriteRead( SEQ_PROGRAM );

                         // write starting address byte by byte, MSB first...
                         spi_data = SPI_WriteRead( flash_start_addr >> 16 );
                         spi_data = SPI_WriteRead( flash_start_addr >> 8 );
                         spi_data = SPI_WriteRead( flash_start_addr >> 0 );

                         sprintf(sbuffer,"\nWrt %ld from [%ld..%ld]",flash_value, 
                                                                     flash_start_addr, 
                                                                     flash_start_addr + flash_count-1);
                                                                     
                         UART_Print_String( sbuffer );                            

                         // write first byte
                         spi_data = SPI_WriteRead( flash_value );

                         // de-assert set CS to SPI select channel 0 (null)
                         SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                         
                         // now write remaining data
                         for (index = 0; index < flash_count-1; index++)
                             {

                             // set CS to SPI select channel 3 (FLASH)
                             SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);
                  
                             // re-issue seq program command, but NO address needed
                             spi_data = SPI_WriteRead( SEQ_PROGRAM );
                             // now data
                             spi_data = SPI_WriteRead( flash_value );

                             // set CS to SPI select channel 0 (null)
                             SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                             // progress bar..    
                             UART_Transmit('.');

                             } // end for index

                        // disable SPI interface

                        // set CS to SPI select channel 0 (null)
                        SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                        // finally send write disable command

                        // set CS to SPI select channel 3 (FLASH)
                        SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);
                        
                        spi_data = SPI_WriteRead( WRITE_DISABLE );
                        
                        // set CS to SPI select channel 0 (null)
                        SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);


                        } break;
#endif

#if 1
                    case BYTE_PROGRAM: 
                        {
                        //RESET_BIT(SPI_PORT, SPI_CS);

                        // set CS to SPI select channel 3 (FLASH)
                        SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                         UART_Print_String("\nExec byte Pgm...");

                         flash_start_addr = atoi2( tokens[ 2 ] );
                         flash_count      = atoi2( tokens[ 3 ] ); 
                         flash_value      = atoi2( tokens[ 4 ] ); 


                         sprintf(sbuffer,"\n%ld %ld %ld",flash_start_addr,
                                                       flash_count,
                                                       flash_value);


                         UART_Print_String( sbuffer );                            

                         // initiate byte program command
                         spi_data = SPI_WriteRead( BYTE_PROGRAM );

                         // write starting address byte by byte, MSB first...
                         spi_data = SPI_WriteRead( flash_start_addr >> 16 );
                         spi_data = SPI_WriteRead( flash_start_addr >> 8 );
                         spi_data = SPI_WriteRead( flash_start_addr >> 0 );

                         sprintf(sbuffer,"\nWrt %ld from [%ld..%ld]",flash_value, 
                                                                     flash_start_addr, 
                                                                     flash_start_addr + flash_count-1);
                                                                     
                         UART_Print_String( sbuffer );                            

                         // write  byte
                         spi_data = SPI_WriteRead( flash_value );

                        // disable SPI interface
                        //SET_BIT(SPI_PORT, SPI_CS);
                        // set CS to SPI select channel 0 (null)
                        SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                        } break;
#endif

                    case BLOCK_ERASE4:     
                        {
                        // get address for block to erase
                        if ( (num_tokens-1) >= next_token )
                          flash_start_addr = atoi2( tokens[ next_token ] );
                        else
                          flash_start_addr = 0; // default to block 0

                        // enable SPI interface
                        //RESET_BIT(SPI_PORT, SPI_CS);
                        // set CS to SPI select channel 3 (FLASH)
                        SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                        UART_Print_String("\nExec block erase 4K Cmd...");
                        spi_data = SPI_WriteRead( BLOCK_ERASE4 );

                        // write block address, there are only a handful of blocks since each is 4-64K
                        spi_data = SPI_WriteRead( flash_start_addr >> 16 );
                        spi_data = SPI_WriteRead( flash_start_addr >> 8 );
                        spi_data = SPI_WriteRead( flash_start_addr >> 0 );

                        // disable SPI interface
                        // set CS to SPI select channel 0 (null)
                        SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                        } break;

#if 1

                    case CHIP_ERASE:    
                        {
                        // enable SPI interface
                        //RESET_BIT(SPI_PORT, SPI_CS);

                        // set CS to SPI select channel 3 (FLASH)
                        SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                        UART_Print_String("\nExec Chip Erase...");
                        spi_data = SPI_WriteRead( CHIP_ERASE );

                        // disable SPI interface
                        //SET_BIT(SPI_PORT, SPI_CS);

                        // set CS to SPI select channel 0 (null)
                        SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                        // enable SPI interface
                        //RESET_BIT(SPI_PORT, SPI_CS);

                        // set CS to SPI select channel 3 (FLASH)
                        SPI_CS_PORT = (1 << SPI_CS0) | (1 << SPI_CS1);

                        // poll spi interface, wait for busy to be 0
                        spi_data = SPI_WriteRead( READ_STATUS );

                        // progress bar counter
                        index=0; 
                
                        do {

                           // progress bar..    
                           UART_Transmit('.');

                           _delay_s(1);

                           } while (SPI_WriteRead( 0xFF ) & 0x01 );

                        // disable SPI interface
                        //SET_BIT(SPI_PORT, SPI_CS);

                        // set CS to SPI select channel 0 (null)
                        SPI_CS_PORT = (0 << SPI_CS0) | (0 << SPI_CS1);

                        } break;
#endif


                    default: 
                        {
                        UART_Print_String( "\nCmd not supported.");
                        } break;
                    } // end switch


                  } // end if FLASH

#endif


#if (COMPILE_HELP_SUPPORT)
               else
               if ( !strcmp( tokens[0], "HELP") )
                  {
                  // enter handler
                  // needs work!!! help needs help :)
                  UART_Print_String("\nCommand List                                                           ");
                  UART_Print_String("\n______________________________________________________________________\n");
                  UART_Print_String("\ncls       - Clears the screen.                                          ");
                  UART_Print_String("\nhelp      - Displays this menu (help needs help!)                       ");

                  } // end if HELP

#endif

#if (COMPILE_PORTIO_SUPPORT)
                /*
                Local AVR Port I/O, allows you to write to port B, C, or D, output only for now, just for playing around and testing.
                Note: you can mess things up with this, so you might want to be careful and only send data to the local AVR ports if you
                know what you're doing.

                OUTPORT [.B|.C|.D] [0..255] - Sends data to port B, C, or D respectively.

                */
               else
               if ( !strcmp( tokens[0], "OUTPORT") || !strcmp( tokens[0], "OP"))
                  {
                  // enter handler
                      
                  // get port B, C, D

                  // write the port
                  if (num_tokens == 3)
                     {
                     // retrieve value
                     arg1 = atoi2( tokens[2] );

                     if ( !strcmp( tokens[1], ".B"))   
                        {
                        // set dir to output always, write data
                        DDRB  = 0xFF; 
                        PORTB = arg1;
                        } // end if
                     else
                     if ( !strcmp( tokens[1], ".C"))   
                        {
                        // set dir to output always, write data
                        DDRC  = 0xFF; 
                        PORTC = arg1;
                        } // end if
                     else
                     if ( !strcmp( tokens[1], ".D"))   
                        {
                        // set dir to output always, write data
                        DDRD = 0xFF; 
                        PORTD = arg1;
                        } // end if
                     
                     } // end
                  else
                     {
                     sprintf(sbuffer, "\nERROR - OUTPORT missing parameter port [A, B, C, D] and/or value [0..255].");
                     UART_Print_String( sbuffer );
                     } // end else

                  } // end if OUTPORT
#endif

               else
                  {
                   sprintf(sbuffer, "\nERROR - \"%s\" - Unknown Command.", tokens[0] );
                   UART_Print_String( sbuffer );
                  } // end else unknown command

               } // end if num_tokens > 0

            } // end if 

        // reset point
        cmd_buff_index = -1;
  
        UART_Print_String("\nREADY>");

        } break;
    
    case ASCII_BS: // back space
        {
        if (cmd_buff_index >=0 )
           {
           cmd_buff[ cmd_buff_index-- ] = 0;
           UART_putc( c );
           } // end if 

        } break;

    default:
        {
        UART_putc( c );
        cmd_buff[++cmd_buff_index] = c;
        } break;

    } // end switch

   
    } // end if


} // end while

#endif // master conditional

// return to caller (like there is a caller, no operating system!)
return(0);

} // end main





