///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2008
// 
// Filename: CHAM_AVR_GUESS_NUMBER_01.c
//
// Original Author: Andre' LaMothe
// 
// Last Modified: 10.20.09
//
// Overview: This program shows the use of the Chameleon AVR UART and serial communication to the PC. You will need to hook
// up the USB serial port cable to the Chameleon AVR and make sure to put the serial select switch in the down position.
// Once you compile and download the program to the Chameleon AVR, then you will need a terminal program
// running on the PC in VT100 mode. Some good serial programs that support a lot of features are:
//
// PuTTY           - http://www.chiark.greenend.org.uk/~sgtatham/putty/
// Zoc             - http://www.emtec.com/zoc/
// Absolute Telnet - http://www.celestialsoftware.net/telnet/
//
// There are many others, but these are pretty good. However, do not, under ANY circumstances use Windows HYPERTERMINAL -
// it is JUNK. Locks up, terriably written, single threaded, bad news.
//
// Terminal Program Setup: Once you have installed a terminal program, you will run it and set it up for communication.
// You will need to connect it to the correct serial port and then set the following settings:
//
// Data Format: "N81", no parity, 8 data bits, 1 stop bit.
// Baud Rate: 115200.
// Handshaking: None.
// Terminal Mode: VT100.
// Screen Size: 80x24 or larger, 120x50 is nice.
// Local Echo: Off.
//
// Once you have the setup, go ahead and place the program in terminal mode, hit RESET on the Chameleon and you should see
// the demo start.
//
// Controls: PC + PC keyboard
// 
// 
// Demo must be linked with following source files (in addition to primary file):
// 
// CHAM_AVR_SYSTEM_V010.c              - System variables.
// CHAM_AVR_UART_DRV_V010.c            - UART driver.
// CHAM_AVR_TWI_SPI_DRV_V010.c         - SPI driver.
// CHAM_AVR_SOUND_DRV_V010.c           - Sound driver.
//
// Note: header files for all sources must be in current working directory, so compiler can find them.
// 
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// __AVR_ATmega328P__ , __AVR_ATmega168__
#define __AVR_ATmega328P__ 

// include everything 
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h> 
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>
#include <avr/wdt.h>

// include local Chameleon API header files
#include "CHAM_AVR_SYSTEM_V010.h"
#include "CHAM_AVR_UART_DRV_V010.h"
#include "CHAM_AVR_SOUND_DRV_V010.h"
#include "CHAM_AVR_TWI_SPI_DRV_V010.h"        // you need this always

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES AND MACROS /////////////////////////////////////////////////////////////////////////////////////////////////
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// define CPU frequency in Mhz here if not defined in Makefile or other includes, compiler will throw a warning, ignore
#ifndef F_CPU
#define F_CPU 16000000UL // 28636360UL, 14318180UL, 21477270 UL
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TYPES/CLASSES //////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GLOBALS ////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

char sbuffer[80]; // string buffer


// .X...X.XXXXX.XXXXX...XXXX..XXXXX.XXX.XXX.XXXX.XXXXX.
// .X...X.X...X.X..XX...X.....X...X.X..X..X.X....X.....
// .X...X.XXXXX.XXXX....X..XX.XXXXX.X.....X.XXX..XXXXX. 
// .X.X.X.X...X.X..X....X...X.X...X.X.....X.X........X. 
// .XXXXX.X...X.X...X...XXXXX.X...X.X.....X.XXXX.XXXXX.

// the title screen bitmap as a matrix of byte, stored in FLASH for fun, then to render we will convert to 
// characters: BLOCK and SPACE
// notice the use of "const" and "PROGMEM", also to access the array, we must use "pgm_read_byte( addr )"
// size of banner 52x5
const unsigned char title_string_flash[] PROGMEM = {

0,1,0,0,0,1,0,1,1,1,1,1,0,1,1,1,1,1,0,0,0,1,1,1,1,0,0,1,1,1,1,1,0,1,1,1,0,1,1,1,0,1,1,1,1,0,1,1,1,1,1,0, 
0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,1,1,0,0,0,1,0,0,0,0,0,1,0,0,0,1,0,1,0,0,1,0,0,1,0,1,0,0,0,0,1,0,0,0,0,0, 
0,1,0,0,0,1,0,1,1,1,1,1,0,1,1,1,1,0,0,0,0,1,0,0,1,1,0,1,1,1,1,1,0,1,0,0,0,0,0,1,0,1,1,1,0,0,1,1,1,1,1,0, 
0,1,0,1,0,1,0,1,0,0,0,1,0,1,0,0,1,0,0,0,0,1,0,0,0,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,1,0,0,0,0,0,0,0,0,1,0, 
0,1,1,1,1,1,0,1,0,0,0,1,0,1,0,0,0,1,0,0,0,1,1,1,1,1,0,1,0,0,0,1,0,1,0,0,0,0,0,1,0,1,1,1,1,0,1,1,1,1,1,0,
};

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// EXTERNALS //////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PROTOTYPES /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FUNCTIONS  /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int Get_String( char *cmd_buff, char **tokens )
{
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// this function waits for the user to input a string on from the UART, it supports some 
// editing such as backspace. When the user hits "return" the string is tokenized into an array
// of strings **tokens. The number of tokens extracted are returned by the function
// INPUTS:
//          *buffer  = pointer to string storage where the raw ASCII string from the UART is placed
//          **tokens = an array of pointers where the function puts the tokenized strings
// 
// OUTPUTS: 
//          returns the number of tokens.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int c;
char token_buff[64];
int num_tokens = 0;
int next_token = 0;
int cmd_buff_index = -1;
char *token_ptr = NULL;
char *result_ptr;

// enter infinite while until valid string is entered
while(1)
{
// process next character
if ( (c = UART_getc()) != -1 )
    {
    // special characters
    switch (c)
    {
    case 0x0D: // carriage return
    case 0x0A: // line feed
        {
        cmd_buff[++cmd_buff_index] = 0;

        // any command to process?
        if (cmd_buff_index > 0)
           {        
            // process command...

            // upcase the string to normalize tokens
            strupr( cmd_buff );

            // copy cmd_buffer to token_buffer for processing
            memcpy(token_buff, cmd_buff, strlen( cmd_buff ) + 1);

            // initialize command processor
            num_tokens = 0;

            // initialize token processor
            token_ptr = strtok_r( token_buff, "=, ", &result_ptr);
           
            while( token_ptr ) 
                 {
                 // insert token into token array and inc number of tokens found
                 tokens[ num_tokens++ ] = token_ptr;   
                 token_ptr = strtok_r( NULL, "=, ", &result_ptr);
                 } // end while

            // string has been entered and tokenized, now return number of tokens
            return( num_tokens );
            } // end if

        // empty string
        return (-1);
    
        } break;

    case 0x08: // back space
         {
         if (cmd_buff_index >=0 )
            {
            UART_putc( c );
            cmd_buff_index--;
            } // end if 
         } break;

    default: // otherwise normal character, insert into command string and echo
        {
        UART_putc( c );
        cmd_buff[++cmd_buff_index] = c;
        } break;

    } // end switch

  } // end if 

} // end while

} // end Get_String


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MAIN ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int main(void)
{
int c;
long index, index2;
int result;

int num_tokens;
char cmd_buff[64];
char *tokens[8];

long arg0, arg1, arg2, arg3, arg4; // generic args for command processor
long number, guess;

// initialize UART
UART_Init( UART_DEFAULT_BAUDRATE );
_delay_ms( 100 );

// enable interrupts
sei();

// initialize and set SPI rate
SPI_Init( SPI_DEFAULT_RATE );

// give Propeller driver time to boot
_delay_ms( 2500 );

// clear the screen and home cursor
Terminal_Clear_Screen();

// print header
UART_Print_String("\nChameleon WarGames (C) 2019\n");
_delay_ms( 2500 );

UART_Print_String("\nentering loop..\n");

// draw "WAR GAMES" with characters from FLASH
for (index = 0; index < 52*5; index++)
    {
    // read the image from FLASH memory with special function call that uses "LPM" ASM instruction to read FLASH
    c = pgm_read_byte( &title_string_flash[ index ] );

    // based on data convert to character and send to UART
    if (c == 0)
       {
       Sound_Play(0, 1000, 1);
       UART_Print_String(" ");
       }
    else
    if (c == 1)
       {
       Sound_Play(0, 2000, 1);
       UART_Print_String("#");
       } // end if

    if ( (index % 52) == 0)
       UART_Print_String("\n");
    
    _delay_ms(10);
    
    } // end for

Sound_Stop(0);

// print some fun text to make it seem like a smart computer
UART_Print_String("\n\nHello Dr. Falkner...\n");
_delay_ms(2000);

UART_Print_String("\nJust a moment while I initialize my systems:");
_delay_ms(1000);

UART_Print_String("\nHolographic Memory:");
Sound_Play(0, 500, 1);
_delay_ms(1000);
UART_Print_String("OK");

UART_Print_String("\nMultiprocessor Sub-Systems:");
Sound_Play(0, 500, 1);
_delay_ms(1000);
UART_Print_String("OK");

UART_Print_String("\nQuantum Telecom Links:");
Sound_Play(0, 500, 1);
_delay_ms(1000);
UART_Print_String("OK");

UART_Print_String("\nLoading Autonomous Systems");
Sound_Play(0, 500, 1);

for (index=0; index < 10; index++)
    {
    UART_Print_String(".");
    _delay_ms(250);
    } // end for

Sound_Play(0, 1000, 3);

UART_Print_String("\n\nThat's much better.\n");

// enter main game loop
while(1)
    {
    UART_Print_String("\nWould you like to play a game (yes or no)?");
 
    // get the string from the UART
    num_tokens = Get_String( cmd_buff, tokens );

    if ( num_tokens > 0 && (!strcmp( tokens[0], "YES") || !strcmp( tokens[0], "Y")) )
       {
       UART_Print_String("\n\nExcellent, let's play 'Guess My Number'...\n"); 
       _delay_ms(1000);
       UART_Print_String("\nI am thinking of a number from 1 to 100, try and guess.\n"); 
       _delay_ms(1000);
       UART_Print_String("\nOk, ready...\n"); 

       // get random number
       number = RAND_RANGE(1, 100);

       // enter guessing loop
       while(1)
            {
            UART_Print_String("\nYour Guess?");

            // get the string from the UART
            num_tokens = Get_String( cmd_buff, tokens );
    
            // convert string to number
            guess = atoi2( tokens[0] );

            // conditional logic
            if (guess > number)
               UART_Print_String("\n\nToo high, your going to have to do better...\n");
            else
            if (guess < number)
               UART_Print_String("\n\nToo low, this is getting embarasing for you...\n");
            else
               {
               UART_Print_String("\n\nThat's correct! Maybe there is hope for carbon based lifeforms!\n");
               break;        
               } // end else

            } // end while
              
       } // end if 
    else
    if (  num_tokens > 0 && (!strcmp( tokens[0], "NO") || !strcmp( tokens[0], "N")) )
       {
       UART_Print_String("\n\nWell fine, I am now running a holographic worm on your computer's memory...");

       // and lock up the Chameleon for fun
       while(1);

       } // end if 
    else
       {
       // user is typing bad commands, let him know!
       switch( RAND_RANGE(0,3) )
            {
            case 0: UART_Print_String("\nI don't understand..please try again.\n"); break;
            case 1: UART_Print_String("\nCan you please answer the question again?\n"); break;
            case 2: UART_Print_String("\nI am starting to loose my patience Dr. Falkner...\n"); break;
            case 3: UART_Print_String("\nI can see why SkyNet wants to exterminate your kind.\n"); break;
            } // end switch
        
       } // end else

    } // end while

// return to caller (like there is a caller, no operating system!)
return(0);

} // end main





