///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// 
// Copyright Nurve Networks LLC 2009
// 
// Filename: CHAM_AVR_UART_DRV_V010.c
//
// Original Author: Andre' LaMothe
// 
// Last Modified: 4.28.09
//
// Description: UART library file
// 
// Overview:
//  
// 
// ANSI/ASCII references: online: http://en.wikipedia.org/wiki/ANSI_color
//                                http://www.termsys.demon.co.uk/vtansi.htm
//                                http://ascii-table.com/ansi-escape-sequences.php
//                  CD-ROM:
//
// I/O pin map for UART interface
// 
// SER_TX | PD1 | Pin 3 | Output
// SER_RX | PD0 | Pin 4 | Input
//
// Note: The UART can also be used as another SPI channel.
//
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// INCLUDES ///////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// this processor include needs to happen in the master file
// #define __AVR_ATmega168__

#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <avr/eeprom.h>
#include <avr/io.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <inttypes.h>
#include <compat/twi.h>
#include <avr/pgmspace.h>

// include local CHAMELEON AVR API header files
#include "CHAM_AVR_SYSTEM_V010.h"
#include "CHAM_AVR_UART_DRV_V010.h"

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// MACROS /////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// DEFINES/////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// TYPES/CLASSES //////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// GLOBALS ////////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// global that holds UART current baud rate
long g_baudrate = UART_DEFAULT_BAUDRATE;

// receiver and transmitter head and tail indices for circular buffers
int g_uart_rx_head = 0;
int g_uart_rx_tail = 0;
int g_uart_tx_head = 0;
int g_uart_tx_tail = 0;

// buffers for uart interrupt driver
UCHAR g_uart_buffer_rx[UART_RX_BUFF_SIZE];    // receiver buffer for incoming characters
UCHAR g_uart_buffer_tx[UART_RX_BUFF_SIZE];    // transmit buffer for outgoing characters

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// EXTERNALS //////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// PROTOTYPES /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// FUNCTIONS  /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// UART FUNCTIONS /////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int UART_Init(long baudrate)
{
// initialize the UART to N81 data format, enables interrupts etc. and computes baudrate register value
// note 1: we do not need to initialize the port pins used by the UART since enabling overrides any settings we might
// assign, but it can't hurt to enable the TX to output and RX to input at some point...

// note 2: due to the overclocking of the AVR at @28.63636, the usual formula to compute the baud rate register isn't 
// accurate enough, therefore, constants are used instead. If you slow the processor back down to 20 Mhz or under then 
// the formula will work again
// baudrate_reg = ((F_CPU)/((baudrate)*16))-1;

int baudrate_reg;

// run this code only for 8*NTSC overclock modes, else try and use general formula
if (F_CPU == 28636360UL)
    {
    // based on sent baud rate compute baudrate register value
    switch( baudrate )
        {
        case 9600: // this range works [176..188]
            {
            baudrate_reg = 182; 
            } break;

        case 19200: // this range works [88..90]
            {
            baudrate_reg = 89; 
            } break;

        case 38400: // this range works [45], 38400 is the absolute limit you can run the overclocked UART at
            {       // if you can't tolerate any errors, drop back to 19200, else 38400 is 99.9% error free
            baudrate_reg = 45; 
            } break;
        
        default:  // caller is requesting a non-computed baud rate, use formula and hope for the best
            {
            baudrate_reg = ((F_CPU)/((baudrate)*16))-1;
            } break;

        } // end switch
    } // end if
else // use formula
    {
    // best guess
    // initialize the UART and set baudrate
    baudrate_reg = ((long)F_CPU / ((long)baudrate*16L) - 0L);

    } // end else

// UBRR0 = baudrate;

// UART baud rate register
UBRR0H = (unsigned char)(baudrate_reg >> 8);
UBRR0L = (unsigned char)(baudrate_reg);

// table of various experimental settings, format: "[ baud rate register ] @clock / baudrate"
// [22] @14.31818 / 38400, [14]  @14.31818 / 57600,  [7] @14.31818 / 115200, 
// [11] @21.4775 / 115200,  [33] @21.4775 / 38400, 
// [45] @28.63636 / 38400, [88..90] @28.63636 / 19200, [176..188] @28.63636 / 9600
// [216] @ 33.3333 / 9600
                        
// assign global
g_baudrate = baudrate;

// UART control register
// enable receiver, transmitter and receiver interrupt
UCSR0B = (1 << RXEN0) | (1 << TXEN0) | (1 << RXCIE0); 

// now set data frame format to N:8:1, async mode
// 168, 644
UCSR0C = (0 << USBS0) | ( 1 << UCSZ01) | (1 << UCSZ00) | (0 << UMSEL00) | (0 << UMSEL01) | (0 << UPM01) | ( 0 << UPM00 );

// 162
//UCSR0C = (0<<USBS0)|(1<<UCSZ01)|(1<<UCSZ00)|(0<<UMSEL0)|(0<<UMSEL1)|(0<<UPM01)|(0<<UPM00);

// return success, return better error code later
return(1);

} // end UART_Init

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int UART_Shutdown(void)
{
// shuts down the UART by disconnecting the interrupts

// UART control register
// disable receiver, transmitter and receiver interrupt
UCSR0B = (0 << RXEN0) | (0 << TXEN0) | (0 << RXCIE0); 

// return success
return(1);

} // end UART_Shutdown

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void UART_Transmit(unsigned char ch)
{
// wait for transmit buffer to empty then write data to data register
while (!(UCSR0A & (1<<UDRE0)));

// write data into transmit register
UDR0 = ch;

} // end UART_Transmit

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int UART_putc(unsigned char ch)
{
// places a character into the transmit buffer
// step 1: insert character into buffer
g_uart_buffer_tx[ g_uart_tx_head ] = ch;

// step 2: compute new head index, make sure we don't have an overrun tail
if ( (++g_uart_tx_head) >= UART_TX_BUFF_SIZE)
    g_uart_tx_head = 0;

// step 3: enable the interrupt
UCSR0B |= (1<<UDRIE0);

// step 4: test for overrun
if ( g_uart_tx_head == g_uart_tx_tail)
    {
    // transmit buffer overflow
    // continue, operating, and overwrite data, another approach is to disallow more data like DOS/CMD shell
    // UART_Transmit( 'X' );

    return(0);
    } // end if

// return the character sent 
return(ch);

} // end UART_putc

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int UART_getc(void)
{
// remove character from receiver buffer, if buffer is empty returns -1
int ch; 

// check if there is a character ready?
if ( g_uart_rx_head != g_uart_rx_tail)
    {
    // read character first
    ch = g_uart_buffer_rx[ g_uart_rx_tail ];

    // wrap index around
    if ( (++g_uart_rx_tail) >= UART_RX_BUFF_SIZE)
        g_uart_rx_tail = 0;

    // return the character
    return ( ch );

    } // end if
else
    return ( -1 ); // buffer empty

} // end UART_getc

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int UART_Receive(int wait)
{
// wait - controls if the function blocks or returns immediately, 1 = wait, 0 = don't wait
// waits for incoming receiver to get a character, blocking
// return character code, or -1 for no character in buffer

if (wait)
   {
   // wait for data in receive buffer
   while (!(UCSR0A & (1<<RXC0)));

   // character ready
   return( UDR0 );
   } // end if wait
else 
if ( !(UCSR0A & (1<<RXC0)) )
   // return data to caller
   return( -1 );
else 
   return( UDR0 );

} // end UART_Receive              

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void UART_Newline(void)
{
UART_Transmit( 0x0D );
UART_Transmit( 0x0A );
} // end UART_Newline

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void UART_Print_String(char *string)
{
while (*string != 0)
  {
  if (*string == 0x0A) // line feed
     {
     UART_Transmit( 0x0A );
     UART_Transmit( 0x0D );
     } // end if
  else
     UART_Transmit(*string);

  string++;

  } // end while

} // end 

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void UART_Print_Int(int num)
{
char sbuffer[16];

sprintf(sbuffer,"%d", num);
UART_Print_String( sbuffer );

} // end UART_Print_Int


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

void Terminal_Clear_Screen(void)
{
// this clears the serial terminal off, homes the cursor using VT100/ANSI codes

// clear
UART_Transmit( 27 );
UART_Transmit( '[' );
UART_Transmit( '2' );
UART_Transmit( 'J' );

// home
UART_Transmit( 27 );
UART_Transmit( '[' );
UART_Transmit( 'H' );

} // end Terminal_Clear_Screen

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// UART ISRS  /////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// USART_RX_vect
ISR(USART_RX_vect)
{
// ISR for UART0 receiver character ready

// insert character into circular buffer

// step 1: insert character into buffer
g_uart_buffer_rx[ g_uart_rx_head ] = UDR0;

// step 2: compute new head index, make sure we don't have an overrun tail
if ( (++g_uart_rx_head) >= UART_RX_BUFF_SIZE)
    g_uart_rx_head = 0;

// step 3: test for overrun
if ( g_uart_rx_head == g_uart_rx_tail)
    {
    // receive buffer overflow
    // continue, operating, and overwrite data, another approach is to disallow more data like DOS/CMD shell
    UART_Transmit( 'X' );

    } // end if

} // end ISR(USART_RX_vect)

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// USART_UDRE_vect
ISR(USART_UDRE_vect)
{
// ISR for UART0 transmitter interrupt, UDRIE, UDR = empty & UDRIE -> calls this interrupt

// is there anything to transmit?
if ( g_uart_tx_head != g_uart_tx_tail)
    {
    // transmit character first
    UDR0 = g_uart_buffer_tx[ g_uart_tx_tail ];

    // wrap index around
    if ( (++g_uart_tx_tail) >= UART_TX_BUFF_SIZE)
        g_uart_tx_tail = 0;

    } // end if
else
    // disable the interrupt by resetting the UDRIE bit
    UCSR0B &= ~(1<<UDRIE0);

} // end ISR(USART_UDRE_vect)

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

// USART_TX_vect
ISR(USART_TX_vect)
{
// ISR for UART0 transmitter interrupt, TXCIE0, transmit complete -> calls this interrupt

// is there anything to transmit?
if ( g_uart_tx_head != g_uart_tx_tail)
    {
    // transmit character first
    UDR0 = g_uart_buffer_tx[ g_uart_tx_tail ];

    // wrap index around
    if ( (++g_uart_tx_tail) >= UART_TX_BUFF_SIZE)
        g_uart_tx_tail = 0;

    } // end if
else
    // disable the interrupt by resetting the TXCIE0 bit
    UCSR0B &= ~(1<<TXCIE0);

} // end ISR(USART_TX_vect)



