/* db_chameleon.c - chameleon intrinsic functions
 *
 * Copyright (c) 2009 by David Michael Betz.  All rights reserved.
 *
 */

#include <stdlib.h>
#include "../runtime/db_vm.h"

#ifdef AVR_COMMON
#include <util/delay.h> 
#include "../chamvmavr/CHAM_AVR_SYSTEM_V010.h"
#include "../chamvmavr/CHAM_AVR_TWI_SPI_DRV_V010.h"			
#endif

#ifdef PIC_COMMON
#define _delay_ms(n)    DELAY_MS(n)
#define _delay_us(n)    DELAY_US(n)
#include "../chamvmpic/CHAM_PIC_SYSTEM_V010.h"
#include "../chamvmpic/CHAM_PIC_I2C_SPI_DRV_V010.h"			
#endif

#define GFX_DISPLAY_WIDTH           32
#define GFX_DISPLAY_HEIGHT          24
#define GFX_TILE_MAP_CNT            4
#define GFX_TILE_WIDTH              8
#define GFX_TILE_HEIGHT             8
#define GFX_TILE_CNT                (144 + 16)
#define GFX_PALETTE_CNT             16

#define TILE_MAP_SIZE_IN_BYTES      (sizeof(uint16_t) * GFX_DISPLAY_WIDTH * GFX_DISPLAY_HEIGHT)
#define TILE_MAP_SPACE_IN_BYTES     (TILE_MAP_SIZE_IN_BYTES * GFX_TILE_MAP_CNT)
#define TILE_SIZE_IN_BYTES          (sizeof(uint16_t) * 8)
#define TILE_SPACE_IN_BYTES         (TILE_SIZE_IN_BYTES * GFX_TILE_CNT)
#define PALETTE_SIZE_IN_BYTES       (sizeof(uint8_t) * 4)
#define PALETTE_SPACE_IN_BYTES      (PALETTE_SIZE_IN_BYTES * GFX_PALETTE_CNT)

#define TILE_MAP_CNT                4

typedef struct {
  int16_t offset;
  int16_t width;
  int16_t height;
} TileMapInfo;

static TileMapInfo tileMaps[TILE_MAP_CNT];
static uint16_t tileMapBase;
static uint16_t tileBase;
static uint16_t paletteBase;

FLASH_SPACE char str_bad_tile_map_number[]       = "bad tile map number: %d";
FLASH_SPACE char str_bad_tile_map_offset[]       = "bad tile map offset: %d";
FLASH_SPACE char str_bad_tile_map_width[]        = "bad tile map width: %d";
FLASH_SPACE char str_bad_tile_map_height[]       = "bad tile map height: %d";
FLASH_SPACE char str_bad_tile_map_position[]     = "bad tile map position: %d, %d";

void VM_fcninit(void)
{
    int i;
    
    _delay_ms(3000);

    for (i = 0; i < TILE_MAP_CNT; ++i) {
        tileMaps[i].offset = i * TILE_MAP_SIZE_IN_BYTES;
        tileMaps[i].width = GFX_DISPLAY_WIDTH;
        tileMaps[i].height = GFX_DISPLAY_HEIGHT;
    }

    SPI_Prop_Send_Cmd(GPU_GFX_TILE_MAP_R, 0, 0);
    tileMapBase = (uint16_t)SPI_Prop_Send_Cmd(READ_CMD, 0, 0);

    SPI_Prop_Send_Cmd(GPU_GFX_BITMAP_R, 0, 0);
    tileBase = (uint16_t)SPI_Prop_Send_Cmd(READ_CMD, 0, 0);

    SPI_Prop_Send_Cmd(GPU_GFX_PALETTE_R, 0, 0);
    paletteBase = (uint16_t)SPI_Prop_Send_Cmd(READ_CMD, 0, 0);
}

void VM_fcnstop(void)
{
}

void fcn_prop(Interpreter *i)
{
    uint8_t cmd = i->sp[0];
    SPI_Prop_Send_Cmd(cmd, 0, 0);
    Drop(i, 2);
}

void fcn_prop_8(Interpreter *i)
{
    uint8_t cmd = i->sp[-1];
    uint8_t data = i->sp[0];
    SPI_Prop_Send_Cmd(cmd, data, 0);
    Drop(i, 3);
}

void fcn_prop_8_8(Interpreter *i)
{
    uint8_t cmd = i->sp[-2];
    uint8_t data = i->sp[-1];
    uint8_t data2 = i->sp[0];
    SPI_Prop_Send_Cmd(cmd, data, data2);
    Drop(i, 4);
}

void fcn_prop_16(Interpreter *i)
{
    uint8_t cmd = i->sp[-1];
    uint16_t data = i->sp[0];
    SPI_Prop_Send_Cmd(cmd, (uint8_t)data, (uint8_t)(data >> 8));
    Drop(i, 3);
}

void fcn_prop_result(Interpreter *i)
{
    i->sp[0] = (int16_t)SPI_Prop_Send_Cmd(READ_CMD, 0, 0);
}

void fcn_define_tile_map(Interpreter *i)
{
    int16_t map = i->sp[-3];
    int16_t offset = i->sp[-2];
    int16_t width = i->sp[-1];
    int16_t height = i->sp[0];

    if (map < 0 || map >= TILE_MAP_CNT)
        Abort(i, str_bad_tile_map_number, map);

    if (offset + TILE_MAP_SIZE_IN_BYTES > TILE_MAP_SPACE_IN_BYTES
    ||  offset + width * height * sizeof(uint16_t) > TILE_MAP_SPACE_IN_BYTES)
        Abort(i, str_bad_tile_map_offset, offset);

    if (width < GFX_DISPLAY_WIDTH)
        Abort(i, str_bad_tile_map_width, width);

    if (height < GFX_DISPLAY_HEIGHT)
        Abort(i, str_bad_tile_map_height, height);

    tileMaps[map].offset = offset;
    tileMaps[map].width = width;
    tileMaps[map].height = height;

    Drop(i, 5);
}

void fcn_display_tile_map(Interpreter *i)
{
    int16_t map = i->sp[-2];
    int16_t x = i->sp[-1];
    int16_t y = i->sp[0];
    uint16_t addr;

    if (map < 0 || map >= TILE_MAP_CNT)
        Abort(i, str_bad_tile_map_number, map);

    addr = tileMaps[map].offset + (y * tileMaps[map].width + x) * sizeof(int16_t);

    if (x < 0 || x >= tileMaps[map].width
    ||  y < 0 || y >= tileMaps[map].height
    ||  addr + TILE_MAP_SIZE_IN_BYTES > TILE_MAP_SPACE_IN_BYTES)
        Abort(i, str_bad_tile_map_position, x, y);

    addr += tileMapBase;

    SPI_Prop_Send_Cmd(GPU_GFX_TILE_MAP_W, (uint8_t)addr, (uint8_t)(addr >> 8));

    Drop(i, 4);
}

void fcn_set_term_tile_map(Interpreter *i)
{
    int16_t map = i->sp[0];
    uint16_t addr;

    if (map < 0 || map >= TILE_MAP_CNT)
        Abort(i, str_bad_tile_map_number, map);

    addr = tileMapBase + tileMaps[map].offset;

    SPI_Prop_Send_Cmd(GPU_GFX_TERM_PTR_W, (uint8_t)addr, (uint8_t)(addr >> 8));

    Drop(i, 2);
}

void fcn_load_tile(Interpreter *i)
{
	int16_t tile = i->sp[-1];
    int16_t *array = &i->variables[i->sp[0]];
    int16_t cnt = array[-2];
    if (tile >= 0 && tile < GFX_TILE_CNT && cnt == 8 && array[-1] == 8) {
        uint16_t addr = tileBase + tile * TILE_SIZE_IN_BYTES;
        SPI_Prop_Send_Cmd(GPU_GFX_SRC_ADDR_W, (uint8_t)addr, (uint8_t)(addr >> 8));
        SPI_Prop_Send_Cmd(GPU_GFX_SET_AUTO_INC_W, 2, 0);
        while (--cnt >= 0) {
            int16_t data = *array++;
            SPI_Prop_Send_Cmd(GPU_GFX_RAM_PORT16_W, (uint8_t)data, (uint8_t)(data >> 8));
        }
    }
    Drop(i, 3);
}

void fcn_load_palette(Interpreter *i)
{
	int16_t palette = i->sp[-1];
    int16_t *array = &i->variables[i->sp[0]];
    uint8_t *barray = (uint8_t *)array;
    int16_t cnt = array[-2];
    if (palette >= 0 && palette < GFX_PALETTE_CNT && cnt == 4 && array[-1] == 4) {
        uint16_t addr = paletteBase + palette * PALETTE_SIZE_IN_BYTES;
        SPI_Prop_Send_Cmd(GPU_GFX_SRC_ADDR_W, (uint8_t)addr, (uint8_t)(addr >> 8));
        SPI_Prop_Send_Cmd(GPU_GFX_SET_AUTO_INC_W, 1, 0);
        while (--cnt >= 0) {
            uint8_t data = *barray++;
            SPI_Prop_Send_Cmd(GPU_GFX_RAM_PORT8_W, data, 0);
        }
    }
    Drop(i, 3);
}

void fcn_write_tile(Interpreter *i)
{
    int16_t map = i->sp[-3];
    int16_t paletteTile = i->sp[-2];
    int16_t x = i->sp[-1];
    int16_t y = i->sp[0];
    if (map >= 0 && map < TILE_MAP_CNT
    &&  x >= 0 && x < tileMaps[map].width
    &&  y >= 0 && y < tileMaps[map].height) {
        uint16_t addr = tileMapBase + tileMaps[map].offset;
        addr += (y * tileMaps[map].width + x) * sizeof(int16_t);
        SPI_Prop_Send_Cmd(GPU_GFX_SRC_ADDR_W, (uint8_t)addr, (uint8_t)(addr >> 8));
        SPI_Prop_Send_Cmd(GPU_GFX_RAM_PORT16_W, (uint8_t)paletteTile, (uint8_t)(paletteTile >> 8));
    }
    Drop(i, 5);
}

void fcn_read_tile(Interpreter *i)
{
    int16_t map = i->sp[-3];
    int16_t paletteTile = i->sp[-2];
    int16_t x = i->sp[-1];
    int16_t y = i->sp[0];
    if (map >= 0 && map < TILE_MAP_CNT
    &&  x >= 0 && x < tileMaps[map].width
    &&  y >= 0 && y < tileMaps[map].height) {
        uint16_t addr = tileMapBase + tileMaps[map].offset;
        addr += (y * tileMaps[map].width + x) * sizeof(int16_t);
        SPI_Prop_Send_Cmd(GPU_GFX_SRC_ADDR_W, (uint8_t)addr, (uint8_t)(addr >> 8));
        SPI_Prop_Send_Cmd(GPU_GFX_RAM_PORT16_R, (uint8_t)paletteTile, (uint8_t)(paletteTile >> 8));
        i->sp[-4] = SPI_Prop_Send_Cmd(READ_CMD, 0, 0);
    }
    Drop(i, 4);
}

void VM_displaychar(int ch)
{
    SPI_Prop_Send_Cmd(GFX_CMD_NTSC_PRINTCHAR, ch, 0);
}

